/*
** This file is part of the C Utility Toolkit (CUTK)
** Copyright (C) 2002-2005 Chris Osgood
** http://www.cutk.org
** 
** This library is free software; you can redistribute it and/or
** modify it under the terms of the GNU Lesser General Public
** License as published by the Free Software Foundation; either
** version 2.1 of the License, or (at your option) any later version.
** 
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
** Lesser General Public License for more details.
** 
** You should have received a copy of the GNU Lesser General Public
** License along with this library; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
** 
** Please report all bugs and problems to "bugs at cutk.org".
*/

/*
** See array.h for more information.
**
** $Id: array.c 13 2005-01-11 22:29:27Z chris $
*/

#include <cutk/array.h>
#include <string.h>

/* Utility macro for getting the element size */
#define aryesize(x) ((*((array_type**)(void*)(x))-1)->esize)

/**
 * Allocate a new empty array.  Normally this function is not called directly.
 * Use the aryalloc() macro instead.  This allows you to specify the element
 * type rather than the element size.  Example: "int* array = aryalloc(int);"
 *
 * @param esize Element size.  This is the size (in bytes) of each element in
 *              the array.  This is normally just the sizeof(ElementType).
 * @return This returns a pointer to the array data (not the array structure)
 *         or NULL on failure.
 */
void* aryalloc_(size_t esize)
{
   array_type* ary = (array_type*)malloc(sizeof(array_type));
   if (ary == NULL)
      return NULL;

   ary->max = 0;
   ary->size = 0;
   ary->esize = esize;

   return &ary->ary;
}

/**
 * Allocate a new array of a specified size.  Normally this function is not
 * called directly.  Use the aryallocsize() macro instead.
 * Example: "int* array = aryallocsize(int, 2000);"
 *
 * @param esize Element size.  This is the size (in bytes) of each element in
 *              the array.  This is normally just the sizeof(ElementType).
 * @param size Number of items in array
 * @return This returns a pointer to the array data (not the array structure)
 *         or NULL on failure.
 */
void* aryallocsize_(size_t esize, size_t size)
{
   array_type* ary = (array_type*)malloc(sizeof(array_type) + esize * size);
   if (ary == NULL)
      return NULL;

   ary->max = size;
   ary->size = size;
   ary->esize = esize;

   return &ary->ary;
}

/**
 * This releases all memory held by an array.
 *
 * @param ary Pointer to an array pointer
 */
void aryfree(void* ary)
{
   free(*(array_type**)ary - 1);
   *(void**)ary = NULL;
}

/**
 * This preallocates memory for a certain maximum number of elements.
 * Note that this does not change the size of the array unless max
 * is less than the current size of the array.
 *
 * @param ary Pointer to an array pointer
 * @param max Number of elements to preallocate
 * @return 0 on success; non-zero on failure
 */
int aryreserve(void* ary, size_t max)
{
   if (arymax(ary) == max)
      return 0;
   else
   {
      void* temp = realloc( *(array_type**)ary - 1,
         sizeof(array_type) + max * aryesize(ary) );

      if (temp == 0)
         return 1;

      *(void**)ary = &((array_type*)temp)->ary;
   }

   arymax(ary) = max;

   if (arymax(ary) < arysize(ary))
      arysize(ary) = arymax(ary);

   return 0;
}

/**
 * This resizes an array.  Note that this does not free any memory
 * held by the array if the size is smaller than the current size.
 *
 * @param ary Pointer to an array pointer
 * @param size New size of array
 * @return 0 on success; non-zero on failure
 */
int aryresize(void* ary, size_t size)
{
   if (size > arymax(ary) && aryreserve(ary, size) != 0)
      return 1;

   arysize(ary) = size;

   return 0;
}

/**
 * This adds and copies a new element to the end of the array.
 *
 * @param ary Pointer to an array pointer
 * @param element This value is copied into the new value at the end of the
 *                array
 * @return 0 on success; non-zero on failure
 */
int aryappend(void* ary, const void* element)
{
   if ((arysize(ary) + 1 > arymax(ary) &&
        aryreserve(ary, arysize(ary) ? arysize(ary) * 2 : 1) != 0) ||
       aryresize(ary, arysize(ary) + 1) != 0)
   {
      return 1;
   }

   memcpy(*(char**)ary + (arysize(ary) - 1) * aryesize(ary),
      element, aryesize(ary));

   return 0;
}

/**
 * Inserts a new element at position "index"
 *
 * @param ary Pointer to an array pointer
 * @param index Element index number to insert before
 * @return 0 on success; non-zero on failure
 */
int aryinsert(void* ary, size_t index)
{
   char* ptr;

   if (index > arysize(ary) || aryresize(ary, arysize(ary) + 1) != 0)
      return 1;
   else if (index == arysize(ary) - 1)
      return 0;

   ptr = *(char**)ary + index * aryesize(ary);
   memmove(ptr + aryesize(ary), ptr,
      *(char**)ary + (arysize(ary) - 1) * aryesize(ary) - ptr);

   return 0;
}

/***
 * This erases an element from the array.  Note that this changes the
 * indicies for all following elements.  Also note that this will not
 * free memory held by the array.  Use aryreserve to free memory.
 *
 * @param ary Pointer to an array pointer
 * @param index Element index number to erase
 * @return 0 on success; non-zero on failure
 */
int aryerase(void* ary, size_t index)
{
   char* ptr;

   if (arysize(ary) == 0 || index >= arysize(ary))
      return 1;

   /* The following resize can not fail so no error checking is done */
   aryresize(ary, arysize(ary) - 1);
   if (index == arysize(ary))
      return 0;

   ptr = *(char**)ary + index * aryesize(ary);
   memmove(ptr, ptr + aryesize(ary),
      *(char**)ary + arysize(ary) * aryesize(ary) - ptr);

   return 0;
}

/***
 * This erases an element from the array.  Note that this changes the
 * indicies for all following elements.  Also note that this will not
 * free memory held by the array.  Use aryreserve to free memory.
 *
 * @param ary Pointer to an array pointer
 * @param index Element index number to erase
 * @return 0 on success; non-zero on failure
 */
int aryfasterase(void* ary, size_t index)
{
   char* ptr;

   if (arysize(ary) == 0 || index >= arysize(ary))
      return 1;

   /* The following resize can not fail so no error checking is done */
   aryresize(ary, arysize(ary) - 1);
   if (index == arysize(ary))
      return 0;

   ptr = *(char**)ary + index * aryesize(ary);
   memmove(ptr, ptr + aryesize(ary),
      *(char**)ary + arysize(ary) * aryesize(ary) - ptr);

   return 0;
}

/**
 * This duplicates an array.  The new array will use the minimum amount of
 * memory necessary to hold all the elements in the array.
 *
 * @param ary Pointer to an array pointer
 * @param i Element index number to erase
 * @return Pointer to new array (same as aryalloc)
 */
void* arydup(const void* ary)
{
   void* temp = aryallocsize_(aryesize(ary), arysize(ary));
   if (temp == NULL)
      return NULL;

   memcpy(temp, *(char**)ary, arysize(ary) * aryesize(ary));

   return temp;
}

/**
 * This copies an array.  This is just a convenient way to do an aryfree()
 * followed by arydup().
 *
 * @param dest Pointer to destination array pointer
 * @param src Pointer to source array pointer
 * @return 0 on success; non-zero on failure
 */
int arycpy(void* dest, const void* src)
{
   aryfree(dest);

   *(void**)dest = arydup(src);
   if (*(void**)dest == NULL)
      return 1;

   return 0;
}

/**
 * Appends the source array to the end of the destination
 *
 * @param dest Pointer to destination array pointer
 * @param src Pointer to source array pointer
 * @return 0 on success; non-zero on failure
 */
int arycat(void* dest, const void* src)
{
   size_t orig_size = arysize(dest);

   if (aryresize(dest, arysize(dest) + arysize(src)) != 0)
      return 1;

   memcpy(*(char**)dest + orig_size * aryesize(dest), *(char**)src,
      arysize(src) * aryesize(src));

   return 0;
}
