/*
** This file is part of the C Utility Toolkit (CUTK)
** Copyright (C) 2002-2005 Chris Osgood
** http://www.cutk.org
** 
** This library is free software; you can redistribute it and/or
** modify it under the terms of the GNU Lesser General Public
** License as published by the Free Software Foundation; either
** version 2.1 of the License, or (at your option) any later version.
** 
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
** Lesser General Public License for more details.
** 
** You should have received a copy of the GNU Lesser General Public
** License along with this library; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
** 
** Please report all bugs and problems to "bugs@cutk.org".
*/

/*
** See dll.h for more information.
**
** $Id: dll.c 18 2005-01-12 03:22:52Z chris $
*/

#include <cutk/dll.h>
#include <cutk/dir.h>
#include <stdlib.h>
#include <string.h>

#ifdef _WIN32
#include <windows.h>
#else
#include <dlfcn.h>
#endif

/**
 * Loads a DLL into memory.
 * @param path File path
 * @return DLL handle or 0 on fail
 */
Dll DllOpen(const char* path)
{
   Dll dll;

   char* path0 = malloc(strlen(path) + 1);
   if (!path0)
      return 0;

   strcpy(path0, path);
   CleanPath(path0);

#ifdef _WIN32
   /* Dual stage load on Windows, otherwise we get a dialog pop-up with bad files */
   dll = LoadLibraryEx(path0, 0, LOAD_LIBRARY_AS_DATAFILE);
   if (dll)
   {
      FreeLibrary(dll);
      dll = LoadLibrary(path0);
   }
#else
   dll = dlopen(path0, RTLD_NOW);
#endif

   free(path0);

   return dll;
}

/**
 * Unloads a DLL from memory.
 * @param dll DLL handle
 * @return 0 on sucess; non-zero on fail
 */
int DllClose(Dll dll)
{
   if (dll)
   {
#ifdef _WIN32
      if (!FreeLibrary(dll))
         return 1;
#else
      if (dlclose(dll))
         return 1;
#endif
   }

   return 0;
}

/**
 * Looks up a symbol in a DLL.
 * @param dll DLL handle
 * @param sym Symbol name
 * @return Pointer to symbol or 0 on fail
 */
void* DllSym(Dll dll, const char* sym)
{
#ifdef _WIN32
   void* ptr = GetProcAddress(dll, sym);
#else
   void* ptr = dlsym(dll, sym);
#endif

   return ptr;
}

