/*
** This file is part of the C Utility Toolkit (CUTK)
** Copyright (C) 2002-2005 Chris Osgood
** http://www.cutk.org
**
** This library is free software; you can redistribute it and/or
** modify it under the terms of the GNU Lesser General Public
** License as published by the Free Software Foundation; either
** version 2.1 of the License, or (at your option) any later version.
**
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
** Lesser General Public License for more details.
**
** You should have received a copy of the GNU Lesser General Public
** License along with this library; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**
** Please report all bugs and problems to "bugs at cutk.org".
*/

/*
** See log.h for more information
**
** $Id: log.c 19 2005-01-12 03:29:50Z chris $
*/

#include <cutk/log.h>

#ifdef _WIN32
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#endif

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <limits.h>

#ifdef _WIN32
#define PATH_MAX MAX_PATH
#endif

static char g_log_filename[PATH_MAX + 1] = "flogf.log";
static char g_log_enabled = 1;
static char g_timestamp_enabled = 1;

/**
 * Sets log filename
 * @param filename Filename
 */
void flogf_filename(const char* filename)
{
   strncpy(g_log_filename, filename, sizeof(g_log_filename));
   g_log_filename[sizeof(g_log_filename) - 1] = 0;
}

/**
 * Disables logging
 */
void flogf_disable(void)
{
   g_log_enabled = 0;
}

/**
 * Enables logging
 */
void flogf_enable(void)
{
   g_log_enabled = 1;
}

/**
 * Disables time stamp
 */
void flogf_disableTime(void)
{
   g_timestamp_enabled = 0;
}

/**
 * Enables time stamp
 */
void flogf_enableTime(void)
{
   g_timestamp_enabled = 1;
}

/**
 * Logs stuff.  Syntax is the same as printf()
 * @param msg Message to log
 * @param ... Variable parameters (like printf)
 */
void flogf(const char* msg, ...)
{
#ifdef _WIN32
   DWORD lastErr = GetLastError(); /* Preserve error state */
#endif

   char *str, last;
   size_t i;
   size_t lf_count;
   char *out_buf, *t_buf;
   FILE* fp = 0;
   time_t t;
   struct tm* time_s;
   va_list args;

   if (!g_log_enabled)
      return;

   if (strlen(g_log_filename) > 0)
      fp = fopen(g_log_filename, "ab");

   if (!fp)
   {
      fp = stderr;
      fprintf(fp, "ERROR: no log file\n");
   }

   lf_count = 1;
   str = strchr(msg, '\n');
   while (str)
   {
      str++;
      lf_count++;
      str = strchr(str, '\n');
   }

   out_buf = calloc(1, strlen(msg) + 3 + lf_count * 16);
   if (!out_buf)
   {
      fprintf(fp, "INTERNAL ERROR in logger: No memory\r\n");
      goto ExitLog;
   }

   time(&t);
   time_s = localtime(&t);

   if (!g_timestamp_enabled)
      out_buf[0] = 0x00;
   else
      sprintf(out_buf, "%.2d/%.2d %.2d:%.2d:%.2d ", 
                       time_s->tm_mon+1, 
                       time_s->tm_mday, 
                       time_s->tm_hour, 
                       time_s->tm_min, 
                       time_s->tm_sec);

   out_buf[15] = 0; /* Enforce the time/date width */

   t_buf = &out_buf[strlen(out_buf)];

   for (i = 0, last = 0; msg[i] != 0x00; i++)
   {
      if (last == '\n' && g_timestamp_enabled)
      {
         sprintf(t_buf, "               ");
         t_buf += 15;
      }

      if (msg[i] == '\n' && last != '\r')
      {
         *t_buf = '\r';
         t_buf++;
      }

      *t_buf = msg[i];
      t_buf++;

      last = msg[i];
   }

   if (last != '\n')
   {
      *t_buf = '\r';
      t_buf++;
      *t_buf = '\n';
      t_buf++;
   }

   va_start(args, msg);
   vfprintf(fp, out_buf, args);
   va_end(args);

ExitLog:
   if (fp == stderr)
      fflush(fp);
   else
      fclose(fp);

   if (out_buf)
      free(out_buf);

#ifdef _WIN32
   SetLastError(lastErr); /* Preserve error state */
#endif
}

