/*
** This file is part of the C Utility Toolkit (CUTK)
** Copyright (C) 2002-2005 Chris Osgood
** http://www.cutk.org
**
** This library is free software; you can redistribute it and/or
** modify it under the terms of the GNU Lesser General Public
** License as published by the Free Software Foundation; either
** version 2.1 of the License, or (at your option) any later version.
**
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
** Lesser General Public License for more details.
**
** You should have received a copy of the GNU Lesser General Public
** License along with this library; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**
** Please report all bugs and problems to "bugs at cutk.org".
*/

/*
** See plugin.h for more information
**
** $Id: plugin.c 30 2005-01-12 16:32:46Z chris $
*/

#include <cutk/plugin.h>
#include <cutk/dir.h>

/**
 * Initializes a new plugin manager
 * @param mgr Pointer to empty PluginManager structure to be filled in
 * @return 0 on success; non-zero on fail
 */
int PluginManagerAlloc(PluginManager* mgr)
{
   mgr->plugins_ = aryalloc(Plugin);
   if (!mgr->plugins_)
   {
      mgr->plugins_ = 0;
      mgr->event_map_ = 0;
      return -1;
   }

   mgr->event_map_ = StrHmapAlloc(0);
   if (!mgr->event_map_)
   {
      aryfree(&mgr->plugins_);
      mgr->plugins_ = 0;
      return -2;
   }

   mgr->callbacks_ = aryalloc(CallbackEntry*);
   if (!mgr->callbacks_)
   {
      aryfree(&mgr->plugins_);
      mgr->plugins_ = 0;
      StrHmapFree(mgr->event_map_);
      mgr->event_map_ = 0;
      return -3;
   }

   /* These allow the plugins to call into the main API */
   mgr->AddPluginObj = PluginManagerAddPluginObj;
   mgr->AddPlugin = PluginManagerAddPlugin;
   mgr->AddDir = PluginManagerAddDir;
   mgr->AddEvent = PluginManagerAddEvent;
   mgr->RemoveEvent = PluginManagerRemoveEvent;
   mgr->RemoveEventName = PluginManagerRemoveEventName;
   mgr->FindEvent = PluginManagerFindEvent;
   mgr->DoEvent = PluginManagerDoEvent;
   mgr->DoEventName = PluginManagerDoEventName;

   return 0;
}

/**
 * Destroys a PluginManager context, freeing all memory and resources. This
 * also calls the finalize() function of all active plugins.
 * @param mgr Pointer to PluginManager structure
 */
void PluginManagerFree(PluginManager* mgr)
{
   /* Iterate and call finalize() on plugins */
   size_t i;

   if (mgr->plugins_)
   {
      for (i = 0; i < arysize(&mgr->plugins_); i++)
      {
         if (mgr->plugins_[i].finalize)
            mgr->plugins_[i].finalize();

         if (mgr->plugins_[i].dll_)
            DllClose(mgr->plugins_[i].dll_);
      }

      aryfree(&mgr->plugins_);
      mgr->plugins_ = 0;
   }

   if (mgr->callbacks_)
   {
      for (i = 0; i < arysize(&mgr->callbacks_); i++)
         aryfree(&mgr->callbacks_[i]);
   
      aryfree(&mgr->callbacks_);
      mgr->callbacks_ = 0;
   }

   if (mgr->event_map_)
   {
      StrHmapFree(mgr->event_map_);
      mgr->event_map_ = 0;
   }
}

/**
 * Destroys a PluginManager context, freeing all memory and resources
 * @param mgr Pointer to PluginManager structure
 */
int PluginManagerAddPluginObj(PluginManager* mgr, Plugin* plugin)
{
   if (aryappend(&mgr->plugins_, plugin))
      return -1;

   if (plugin->initialize(mgr))
   {
      aryresize(&mgr->plugins_, arysize(&mgr->plugins_) - 1);
      return -2;
   }

   return 0;
}

/**
 * Adds a plugin DLL to the manager.
 * @param mgr Pointer to PluginManager structure
 * @param path Filesystem path to DLL that will be loaded
 * @return 0 on success; non-zero on fail
 */
int PluginManagerAddPlugin(PluginManager* mgr, const char* path)
{
   int rv = 0;
   void *init, *final;
   Plugin plugin;

   Dll dll = DllOpen(path);
   if (!dll)
   {
      rv = -1;
      goto exitfn;
   }

   init = DllSym(dll, "plugin_init");
   if (!init)
   {
      rv = -2;
      goto exitfn;
   }

   final = DllSym(dll, "plugin_final");
   if (!final)
   {
      rv = -3;
      goto exitfn;
   }

   plugin.dll_ = dll;
   strncpy(plugin.path_, path, sizeof(plugin.path_) - 1);
   plugin.path_[sizeof(plugin.path_) - 1] = 0x00;
   plugin.initialize = (int(*)(struct PluginManager*))init;
   plugin.finalize = (int(*)())final;

   if (PluginManagerAddPluginObj(mgr, &plugin))
   {
      rv = -4;
      goto exitfn;
   }

   /* Keep the DLL open if all is OK */
   dll = 0;

exitfn:
   if (dll)
      DllClose(dll);

   return rv;
}

/**
 * Adds an entire directory of plugins.  This attempts to load all files in
 * the specified directory.  Some may fail if they are not plugins which is
 * not a problem.
 * @param mgr Pointer to PluginManager structure
 * @param path Filesystem path directory
 * @param recursive Set to 0 to not traverse down sub-directories
 * @return 0 on success; non-zero on fail
 */
int PluginManagerAddDir(PluginManager* mgr, const char* path, char recursive)
{
   char* newpath;
   struct dirent* ent;
   DIR* dir = opendir(path);
   if (!dir)
      return -1;

   while ((ent = readdir(dir)))
   {
      if (ent->d_name[0] == '.')
         continue;

      newpath = malloc(strlen(path) + strlen(ent->d_name) + 2);
      if (!newpath)
         return -2;

      strcpy(newpath, path);
      strcat(newpath, "/");
      strcat(newpath, ent->d_name);
      CleanPath(newpath);

      if (recursive && ent->d_type == DT_DIR)
         PluginManagerAddDir(mgr, newpath, 1); /* don't care if it fails */
      else
         PluginManagerAddPlugin(mgr, newpath); /* don't care if it fails */

      free(newpath);
   }

   return 0;
}

/**
 * Adds a named event to the manager.
 * @param mgr Pointer to PluginManager structure
 * @param event Event name
 * @param cb Callback function to call when this event occurs
 * @param pri Priority of event
 * @return The return value is the event ID number which can be used in future
 *         calls to PluginManagerDoEvent() or 0 if AddEvent fails.
 */
size_t PluginManagerAddEvent(PluginManager* mgr, const char* event, PluginCallback cb, unsigned char pri)
{
   size_t index = (size_t)StrHmapFind(mgr->event_map_, event);
   CallbackEntry entry;
   size_t i;

   entry.callback = cb;
   entry.priority = pri;

   if (!index)
   {
      CallbackEntry* callback_array = aryalloc(CallbackEntry);
      if (!callback_array)
         return 0;

      if (aryappend(&mgr->callbacks_, &callback_array))
      {
         aryfree(&callback_array);
         return 0;
      }

      index = arysize(&mgr->callbacks_);

      if (StrHmapInsert(mgr->event_map_, event, (void*)index))
      {
         aryfree(&callback_array);
         aryresize(&mgr->callbacks_, arysize(&mgr->callbacks_) - 1);
         return 0;
      }
   }

   /* Insert based on priority */
   index--;
   for (i = 0; i < arysize(&mgr->callbacks_[index]); i++)
   {
      if (entry.priority < mgr->callbacks_[index][i].priority)
      {
         aryinsert(&mgr->callbacks_[index], i);
         mgr->callbacks_[index][i] = entry;
         goto exitfn;
      }
   }

   aryappend(&mgr->callbacks_[index], &entry);

exitfn:
   return index + 1;
}

/**
 * Removes an event from the manager
 * @param mgr Pointer to PluginManager structure
 * @param event Event ID num
 * @param cb Callback function
 * @param pri Priority of event
 * @return 0 on sucess; non-zero on fail
 */
int PluginManagerRemoveEvent(PluginManager* mgr, size_t event, PluginCallback cb, unsigned char pri)
{
   int rv = -1;
   CallbackEntry* callbacks;
   size_t i;

   if (!event || event > arysize(&mgr->callbacks_))
      return -2;

   callbacks = mgr->callbacks_[event - 1];
   if (!callbacks)
      return -3;

   for (i = 0; i < arysize(&callbacks); i++)
   {
      if (callbacks[i].callback == cb && callbacks[i].priority == pri)
      {
         if (aryerase(&callbacks, i))
            return -4;

         rv = 0;
         i--;
      }
   }

   return rv;
}

/**
 * Removes a named event from the manager.  Slower than RemoveEvent because
 * this does a name lookup.
 * @param mgr Pointer to PluginManager structure
 * @param event Event name
 * @param cb Callback function
 * @param pri Priority of event
 * @return 0 on sucess; non-zero on fail
 */
int PluginManagerRemoveEventName(PluginManager* mgr, const char* event, PluginCallback cb, unsigned char pri)
{
   return PluginManagerRemoveEvent(mgr, (size_t)StrHmapFind(mgr->event_map_, event), cb, pri);
}

/**
 * Finds a named event and returns the ID number.
 * @param mgr Pointer to PluginManager structure
 * @param event Event name
 * @return The return value is the event ID number or 0 if find fails.
 */
size_t PluginManagerFindEvent(PluginManager* mgr, const char* event)
{
   return (size_t)StrHmapFind(mgr->event_map_, event);
}

/**
 * Causes an event to run.  This will call all functions registered for the
 * specified event.
 * @param mgr Pointer to PluginManager structure
 * @param event Event ID number
 * @param param1 Parameter to pass to plugin
 * @param param2 Extra parameter to pass to plugin
 * @return 0 on success; non-zero on failure
 */
int PluginManagerDoEvent(PluginManager* mgr, size_t event, void* param1, void* param2)
{
   CallbackEntry* callbacks;
   PluginResult action;
   size_t i;

   if (!event || event > arysize(&mgr->callbacks_))
      return -1;

   callbacks = mgr->callbacks_[event - 1];
   if (!callbacks)
      return -2;

   for (i = 0; i < arysize(&callbacks); i++)
   {
      action = callbacks[i].callback(event, param1, param2);

      if (action == PLUGIN_STOP)
         break;
      else if (action == PLUGIN_AGAIN)
         i--;
   }

   return 0;
}

/**
 * Causes an event to run.  This will call all functions registered for the
 * specified event.  This is slower than the plain DoEvent becuase it does
 * a lookup of the event name.
 * @param mgr Pointer to PluginManager structure
 * @param event Event name
 * @param param1 Parameter to pass to plugin
 * @param param2 Extra parameter to pass to plugin
 * @return 0 on success; non-zero on failure
 */
int PluginManagerDoEventName(PluginManager* mgr, const char* event, void* param1, void* param2)
{
   return PluginManagerDoEvent(mgr, PluginManagerFindEvent(mgr, event), param1, param2);
}

