/*
** This file is part of the C Utility Toolkit (CUTK)
** Copyright (C) 2002-2005 Chris Osgood
** http://www.cutk.org
**
** This library is free software; you can redistribute it and/or
** modify it under the terms of the GNU Lesser General Public
** License as published by the Free Software Foundation; either
** version 2.1 of the License, or (at your option) any later version.
**
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
** Lesser General Public License for more details.
**
** You should have received a copy of the GNU Lesser General Public
** License along with this library; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**
** Please report all bugs and problems to "bugs at cutk.org".
*/

/*
** See StrHmap.h for more information.
**
** $Id: strhmap.c 8 2005-01-02 16:25:13Z chris $
*/

/* Inline */
#ifdef _MSC_VER
#define INLINE_X static __inline
#else
#define INLINE_X static inline
#endif

#ifdef NO_INLINE
#define INLINE static
#else
#define INLINE INLINE_X
#endif

#include <cutk/strhmap.h>
#include <stdlib.h>

#ifndef SIZE_MAX
#define SIZE_MAX ((size_t)-1)
#endif

/* Some prime values to use when sizing the internal hash table */
static const size_t hash_primelist[] =
{
   53u, 97u, 193u, 389u, 769u, 1543u, 3079u, 6151u, 12289u, 24593u,
   49157u, 98317u, 196613u, 393241u, 786433u, 1572869u, 3145739u,
   6291469u, 12582917u, 25165843u, 50331653u, 100663319u, 201326611u,
   402653189u, 805306457u, 1610612741u, 3221225473u, 4294967291u
};

/* Returns the next prime larger than the specified size */
INLINE size_t NextPrime(size_t size)
{
   size_t i;

   for(i = 0; hash_primelist[i] < size &&
              i < sizeof(hash_primelist) / sizeof(size_t); i++);

   return hash_primelist[i];
}

/* Fast string hashing routine */
INLINE size_t StrHash(const StrHmap* map, const char* s)
{
   size_t i;
   for (i = 0; *s; s++)
      i = 31 * i + *s;
   return i % arysize(&map->hash_tbl);
}

/* Fast string equality test.  In most cases this is faster than strcmp */
INLINE char StrEq(const char* s1, const char* s2)
{
   if (s1 == s2) return 1;
   if (!s1 || !s2) return 0;
   while (*s1 != '\0' && *s1 == *s2)
      s1++, s2++;
   return *s1 == *s2;
}

/* Internal function used by StrHmapInsert and StrHmapReplace */
INLINE int InsertNew(StrHmap* hashmap, const char* key, const void* item, size_t hash)
{
   size_t newitemnum;
   StrHmap_itemtype newitem;

   newitem.key = key;
   newitem.item = item;

   if (hashmap->deleted != SIZE_MAX)
   {
      newitemnum = hashmap->deleted;
      hashmap->deleted = hashmap->items[newitemnum].next;
      hashmap->items[newitemnum] = newitem;
   }
   else
   {
      newitemnum = arysize(&hashmap->items);
      if (aryappend(&hashmap->items, &newitem))
         return 1;
   }

   hashmap->items[newitemnum].next = hashmap->hash_tbl[hash];
   hashmap->hash_tbl[hash] = newitemnum;

    /* For the following it's not really an insert error if rehash fails
       so we don't return an error if it does.  However, the hash table will
       become slow if rehash keeps failing.  Not sure how to notify the
       application of such a condition.  Other more serious errors will
       probably be happening if rehash fails (eg. out of memory). */
   if (arysize(&hashmap->items) > arysize(&hashmap->hash_tbl))
      StrHmapReserve(hashmap, arysize(&hashmap->items)); 

   return 0;   
}

/* External API */

/**
 * Allocate a new empty hash map.
 * @param size Specifies intial expected number of items in the hash map (can
 *             be 0).
 * @return Returns a pointer to the new hash map or NULL on failure.
 */
StrHmap* StrHmapAlloc(size_t size)
{
   StrHmap* hashmap = (StrHmap*)malloc(sizeof(StrHmap));
   if (hashmap == NULL)
      return NULL;

   hashmap->items = aryalloc(StrHmap_itemtype);
   if (hashmap->items == NULL)
   {
      free(hashmap);
      return NULL;
   }
   
   hashmap->hash_tbl = aryallocsize(size_t, NextPrime(size));

   if (hashmap->hash_tbl != NULL)
      memset(hashmap->hash_tbl, 0xFF,
             arysize(&hashmap->hash_tbl) * sizeof(size_t));

   if (hashmap->hash_tbl == NULL)
   {
      aryfree(&hashmap->items);
      free(hashmap);
      return NULL;
   }

   hashmap->deleted = SIZE_MAX;
   
   return hashmap;
}

/**
 * Releases all memory held by a hash map.
 * @param hashmap Pointer to a hash map
 */
void StrHmapFree(StrHmap* hashmap)
{
   aryfree(&hashmap->hash_tbl);
   aryfree(&hashmap->items);
   hashmap->hash_tbl = NULL;
   hashmap->items = NULL;
   free(hashmap);
}

/**
 * Reallocates a hash map.  If the size is smaller than the actual
 * number of elements in the hash map then the effect will be to reduce
 * memory usage if possible.  The size can be larger than the actual number
 * of elements in the hash map which will allow lots of insertions to go
 * faster since the table won't need to be resized as often.
 * @param hashmap Pointer to hash map
 * @param size Number of elements to preallocate
 * @return 0 on success; non-zero on failure
 */
int StrHmapReserve(StrHmap* hashmap, size_t size)
{
   size_t i;
   size_t hash;

   if (size <= arysize(&hashmap->hash_tbl))
      return 0;

   if (aryresize(&hashmap->hash_tbl, NextPrime(size)))
      return 1;

   memset(hashmap->hash_tbl, 0xFF,
          arysize(&hashmap->hash_tbl) * sizeof(size_t));

   for (i = 0; i < arysize(&hashmap->items); i++)
   {
      if (hashmap->items[i].key != NULL)
      {
         hash = StrHash(hashmap, hashmap->items[i].key);
         hashmap->items[i].next = hashmap->hash_tbl[hash];
         hashmap->hash_tbl[hash] = i;
      }
   }

   return 0;
}

/**
 * Finds an element in the hash map.
 * @param hashmap Pointer to hash map
 * @param key Key to search for
 * @return Pointer to element in hash map or NULL if key not found
 */
void* StrHmapFind(StrHmap* hashmap, const char* key)
{
   size_t hash = StrHash(hashmap, key);
   size_t itemnum = hashmap->hash_tbl[hash];

   for (; itemnum != SIZE_MAX; itemnum = hashmap->items[itemnum].next)
   {
      if (StrEq(hashmap->items[itemnum].key, key))
         return (void*)hashmap->items[itemnum].item;
   }

   return NULL;
}

/**
 * Inserts a new element into the hash map.  This will return an error if the
 * key already exists.
 * @param hashmap Pointer to hash map
 * @param key Key to insert element under
 * @param item Item pointer to insert into hash map
 * @return 0 on success; non-zero on failure
 */
int StrHmapInsert(StrHmap* hashmap, const char* key, const void* item)
{
   size_t hash = StrHash(hashmap, key);
   size_t itemnum = hashmap->hash_tbl[hash];

   for (; itemnum != SIZE_MAX; itemnum = hashmap->items[itemnum].next)
   {
      if (StrEq(hashmap->items[itemnum].key, key))
         return 1; /* key exists */
   }

   return InsertNew(hashmap, key, item, hash);
}

/**
 * Inserts or replaces an element in the hash map.  This will overwrite 
 * existing items in the hash map.
 * @param hashmap Pointer to hash map
 * @param key Key to insert element under
 * @param item Item pointer to insert/replace in hash map
 * @return 0 on success; non-zero on failure
 */
int StrHmapReplace(StrHmap* hashmap, const char* key, const void* item)
{
   size_t hash = StrHash(hashmap, key);
   size_t itemnum = hashmap->hash_tbl[hash];

   for (; itemnum != SIZE_MAX; itemnum = hashmap->items[itemnum].next)
   {
      if (StrEq(hashmap->items[itemnum].key, key))
      {
         hashmap->items[itemnum].item = item;
         return 0;
      }
   }

   return InsertNew(hashmap, key, item, hash);
}

/**
 * Erases an element from the hash map.  Note that this will not free
 * memory held by the hash map.  Use StrHmaprehash to free up memory after
 * items have been deleted.
 * @param hashmap Pointer to hash map
 * @param key Key to erase
 * @return 0 on success; non-zero on failure
 */
int StrHmapErase(StrHmap* hashmap, const char* key)
{
   size_t hash = StrHash(hashmap, key);
   size_t itemnum = hashmap->hash_tbl[hash];

   for (; itemnum != SIZE_MAX; itemnum = hashmap->items[itemnum].next)
   {
      if (StrEq(hashmap->items[itemnum].key, key))
      {
         hashmap->items[itemnum].key = NULL;
         hashmap->items[itemnum].item = NULL;
         hashmap->items[itemnum].next = hashmap->deleted;
         hashmap->deleted = itemnum;
         return 0;
      }
   }

   return 1;
}

/**
 * Duplicates a hash map.
 * @param hashmap Pointer to hash map
 * @return Pointer to new hash map
 */
StrHmap* StrHmapDup(const StrHmap* hashmap)
{
   StrHmap* temp = StrHmapAlloc(0);
   if (temp == NULL)
      return NULL;

   if (aryresize(&temp->hash_tbl, arysize(&hashmap->hash_tbl)) ||
       aryresize(&temp->items, arysize(&hashmap->items)))
   {
      StrHmapFree(temp);
      return NULL;
   }

   memcpy(temp->hash_tbl, hashmap->hash_tbl, 
          arysize(&temp->hash_tbl) * sizeof(size_t));

   memcpy(temp->items, hashmap->items,
          arysize(&temp->items) * sizeof(StrHmap_itemtype));

   return temp;
}

/**
 * Rebuilds a hash map.  Erased items are removed from the hash map to
 * free up memory.
 * @param hashmap Pointer to hash map
 * @return 0 on success; non-zero on failure
 */
int StrHmapCompact(StrHmap* hashmap)
{
   size_t i, offset = 0, delcnt = 0;

   /* Compact the items array */
   for (i = 0; offset < arysize(&hashmap->items); i++, offset++)
   {
      while (hashmap->items[offset].key == NULL)
      {
         delcnt++;
         offset++;
      }

      hashmap->items[i] = hashmap->items[offset];
   }

   if (delcnt == 0 && arysize(&hashmap->items) != 0)
      return 0;

   if (aryreserve(&hashmap->items, arysize(&hashmap->items) - delcnt))
      return 1;

   if (aryreserve(&hashmap->hash_tbl, NextPrime(arysize(&hashmap->items))))
      return 1;

   aryclear(&hashmap->hash_tbl);

   /* Rehash all the items */
   return StrHmapReserve(hashmap, arymax(&hashmap->hash_tbl));
}

/**
 * Erases all elements in a hash map.
 * @param hashmap Pointer to hash map
 */
void  StrHmapClear(StrHmap* hashmap)
{
   aryclear(&hashmap->items);
   memset(hashmap->hash_tbl, 0xFF,
          arysize(&hashmap->hash_tbl) * sizeof(size_t));
}
