/*
** This file is part of the C Utility Toolkit (CUTK)
** Copyright (C) 2002-2005 Chris Osgood
** http://www.cutk.org
**
** This library is free software; you can redistribute it and/or
** modify it under the terms of the GNU Lesser General Public
** License as published by the Free Software Foundation; either
** version 2.1 of the License, or (at your option) any later version.
**
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
** Lesser General Public License for more details.
**
** You should have received a copy of the GNU Lesser General Public
** License along with this library; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**
** Please report all bugs and problems to "bugs at cutk.org".
*/

/*
** See timer.h for more information
**
** $Id: timer.c 7 2005-01-02 01:02:22Z chris $
*/

#include <cutk/timer.h>

#ifdef _WIN32
#include <windows.h>
#else
#include <sys/time.h>
#endif

/**
 * Internal function used to update elapsed information
 * @param t Timer object
 */
static void TimerUpdateElapsed(Timer* t)
{
#ifdef _WIN32
   LARGE_INTEGER current;
   QueryPerformanceCounter(&current);
   t->elapsed.QuadPart = current.QuadPart - t->tm.QuadPart;
#else
   struct timeval current;
   gettimeofday(&current, 0);
   t->elapsed = (current.tv_sec - t->tm.tv_sec) * t->freq + current.tv_usec - t->tm.tv_usec;
#endif
}

/**
 * Starts a timer counting
 * @param t Timer object; will be initialized
 */
void TimerStart(Timer* t)
{
#ifdef _WIN32
   t->elapsed.QuadPart = 0;
   QueryPerformanceFrequency(&t->freq);
   t->tm.QuadPart = 0;
   QueryPerformanceCounter(&t->tm);
#else
   t->elapsed = 0;
   t->freq = 1000000;
   t->tm.tv_sec = 0;
   t->tm.tv_usec = 0;
   gettimeofday(&t->tm, 0);
#endif
   t->stopped = 0;
}

/**
 * Returns the elasped time
 * @param t Timer object; will be initialized
 * @param mode Mode to return time in (see TimerMode)
 */
double TimerElapsed(Timer* t, TimerMode mode)
{
   if (!t->stopped)
      TimerUpdateElapsed(t);

#ifdef _WIN32
   if (mode == TIMER_SECS)
      return (double)t->elapsed.QuadPart / (double)t->freq.QuadPart;
   else if (mode == TIMER_FPS)
   {
      if (t->elapsed.QuadPart == 0)
         return 0.0;
      else
         return (double)t->freq.QuadPart / (double)t->elapsed.QuadPart;
   }
   else
      return (double)t->elapsed.QuadPart;
#else
   if (mode == TIMER_SECS)
      return (double)t->elapsed / (double)t->freq;
   else if (mode == TIMER_FPS)
   {
      if (t->elapsed == 0)
         return 0.0;
      else
         return (double)t->freq / (double)t->elapsed;
   }
   else
      return (double)t->elapsed;
#endif
}

/**
 * Stops the timer and returns the elapsed time
 * @param t Timer object
 * @param mode Mode to return time in (see TimerMode)
 */
double TimerEnd(Timer* t, TimerMode mode)
{
   TimerUpdateElapsed(t);
   t->stopped = 1;
   return TimerElapsed(t, mode);
}

