/*
** This file is part of the C Utility Toolkit (CUTK)
** Copyright (C) 2002-2005 Chris Osgood
** http://www.cutk.org
**
** This library is free software; you can redistribute it and/or
** modify it under the terms of the GNU Lesser General Public
** License as published by the Free Software Foundation; either
** version 2.1 of the License, or (at your option) any later version.
**
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
** Lesser General Public License for more details.
**
** You should have received a copy of the GNU Lesser General Public
** License along with this library; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**
** Please report all bugs and problems to "bugs at cutk.org".
*/

/*
** See util.h for more information
**
** $Id: util.c 21 2005-01-12 04:28:32Z chris $
*/

#include <cutk/util.h>

#ifndef __USE_GNU
#define __USE_GNU
#include <string.h>
#undef __USE_GNU
#else
#include <string.h>
#endif

/**
 * Converts a character to its binary equivalent in string form.
 * For example: 0xFF would convert to the string "11111111".
 * @param out  output string (needs space for 8 characters)
 * @param c    binary byte to convert
 */
void Byte2Binary(char* out, unsigned char c)
{
   int i, j;
   for (i = 0, j = 7; i < 8; i++, j--)
      out[j] = ((c >> i) & 0x01) + '0';
}

/**
 * Converts a character to its hex equivalent in string form.
 * For example: 0xFF would convert to the string "FF".
 * @param out  output string (needs space for 2 characters)
 * @param c    binary byte to convert
 */
static char hex_table[] = "0123456789ABCDEF";
void Byte2Hex(char* out, unsigned char c)
{
   out[0] = hex_table[(c & 0xF0) >> 4];
   out[1] = hex_table[c & 0x0F];
}

#if defined(_WIN32) || defined(__APPLE__)
static size_t strnlen(const char* s, size_t max)
{
   size_t i;
   for (i = 0; s[i] != 0x00 && i < max; i++);
   return i;
}
#endif

/**
 * Converts a binary character string to binary
 * For example: "11111111" would convert to the binary value 0xFF.
 * @param in   input string (8 characters or less)
 * @return binary char
 */
unsigned char Binary2Byte(const char* in)
{
   unsigned char i;
   unsigned char len = (unsigned char)strnlen(in, 8);
   unsigned char out = 0;

   for (i = 0; i < len; i++)
      out += (in[len - i - 1] - '0') << i;

   return out;
}

/**
 * Converts a hex string to its binary equivalent
 * For example: "FF" would convert to  binary 0xFF.
 * @param in input string to convert
 * @return binary value
 */
unsigned char Hex2Byte(const char* in)
{
   unsigned char out;

   if (in[0] < 'A')
      out = (in[0] - '0') << 4;
   else if (in[0] < 'a')
      out = (in[0] - 55) << 4;
   else
      out = (in[0] - 87) << 4;

   if (in[1] < 'A')
      out += in[1] - '0';
   else if (in[1] < 'a')
      out += in[1] - 55;
   else
      out += in[1] - 87;

   return out;
}

/**
 * Converts binary data to a hex string
 * @param in   input data to convert
 * @param len  length of input data
 * @return string; must be free()'d to prevent memory leak
 */
char* Bin2Hex(const unsigned char* in, size_t len)
{
   size_t i;
   char* out = (char*)malloc(len * 3);

   if (out == 0 || in == 0 || len == 0)
      return 0;

   for (i = 0; i < len; i++)
   {
      Byte2Hex(&out[i * 3], in[i]);
      out[i * 3 + 2] = 0x20;
   }

   out[len * 3 - 1] = 0x00;

   return out;
}

/**
 * Converts hex string to binary data.
 * Output length is always (unsigned int)(strlen(in)+1 divided by width)
 * @param in      input string to convert
 * @param width   input string character width
 *                ("00 F1 FF " width 3, or "00F1FF" width 2)
 * @return binary data; must be free()'d to prevent memory leak
 */
unsigned char* Hex2Bin(const char* in, size_t width)
{
   size_t i;
   size_t len;
   unsigned char* out;

   if (in == 0 || width == 0)
      return 0;

   len = (strlen(in) + 1) / width; /* FIXME: strlen unsafe! */

   out = (unsigned char*)malloc(len);

   if (out == 0)
      return 0;

   for (i = 0; i < len; i++)
      out[i] = Hex2Byte(&in[i * width]);

   return out;
}
